import contextlib
import json
import time
from typing import Any

from openai import OpenAI


def chat_completion(
    messages: list[dict[str, Any]],
    model: str,
    client: OpenAI,
    tools: list[dict[str, Any]] | None = None,
    tool_choice: str = "auto",
    max_tokens: int | None = None,
    temperature: float = 0.7,
    json_mode: bool = False,
    seed: int | None = None,
    logprobs: tuple[bool, int | None] | None = None,
    n: int = 1,
    **kwargs: Any,
) -> dict[str, Any]:
    """Get an OpenAI chat completion response: https://platform.openai.com/docs/api-reference/chat/create

    NOTE: Depending on the model, certain parameters may not be supported,
    particularly for older vision-enabled models like gpt-4-1106-vision-preview.
    Be sure to check the documentation: https://platform.openai.com/docs/models/gpt-4-turbo-and-gpt-4

    Args:
        messages (list): A list of messages comprising the conversation so far.
        model (str): ID of the model to use. See the model endpoint compatibility table:
            https://platform.openai.com/docs/models/model-endpoint-compatibility
            for details on which models work with the Chat API.
        client (OpenAI): An instance of the OpenAI client.
        tools (list[dict[str, Any]], optional): A list of tools the model may generate JSON inputs for.
            Defaults to None.
        tool_choice (str, optional): The tool choice to use. Can be "auto", "required", "none", or a specific function name.
            Note the function name cannot be any of "auto", "required", or "none". Defaults to "auto".
        max_tokens (int, optional): The maximum number of tokens to generate in the chat completion.
            Defaults to None, which automatically limits to the model's maximum context length.
        temperature (float, optional): What sampling temperature to use, between 0 and 2.
            Higher values like 0.8 will make the output more random,
            while lower values like 0.2 will make it more focused and deterministic. Defaults to 0.7.
        json_mode (bool, optional): When JSON mode is enabled, the model is constrained to only
            generate strings that parse into valid JSON object and will return a dictionary.
            See https://platform.openai.com/docs/guides/text-generation/json-mode
        seed (int, optional): If specified, OpenAI will make a best effort to sample deterministically,
            such that repeated requests with the same `seed` and parameters should return the same result.
            Determinism is not guaranteed, and you should refer to the `system_fingerprint` response
            parameter to monitor changes in the backend.
        logprobs (tuple[bool, int], optional): Whether to return log probabilities of the output tokens or not.
            If `logprobs[0]` is true, returns the log probabilities of each output token returned in the content of message.
            `logprobs[1]` is an integer between 0 and 5 specifying the number of most likely tokens to return at each token position,
            each with an associated log probability. `logprobs[0]` must be set to true if this parameter is used.
        n (int, optional): How many chat completion choices to generate for each input message.
            Defaults to 1.
        **kwargs: Additional keyword arguments to pass to the OpenAI client chat completion.

    Returns:
        dict[str, Any]: A dictionary with the following keys:
            'finish_reason' (str): The reason the model stopped generating further tokens.
                Can be 'stop', 'length', or 'tool_calls'.
            'tool_names' (list[str], optional): The names of the tools called by the model.
            'tool_args_list' (list[dict], optional): The arguments of the tools called by the model.
            'message' (str | dict): The content of the generated assistant message.
                If json_mode is True, this will be a dictionary.
            'logprobs' (list[dict[str, Any] | list[dict[str, Any]]]): If logprobs[1] is between 1 and 5, each element in the list
                will be a list of dictionaries containing the token, logprob, and bytes for the top `logprobs[1]` logprobs. Otherwise,
                this will be a list of dictionaries containing the token, logprob, and bytes for each token in the message.
            'choices' (list[dict], optional): A list of chat completion choices if n > 1 where each dict contains the above fields.
            'completion_tokens' (int): The number of tokens used by the model to generate the completion.
                NOTE: If n > 1 this is the sum of all completions.
            'prompt_tokens' (int): The number of tokens in the messages sent to the model.
            'system_fingerprint' (str, optional): If seed is set, a unique identifier for the model used to generate the response.
            'response_duration' (float): The time, in seconds, taken to generate the response from the API.
    """
    response_format = {"type": "json_object"} if json_mode else None

    kwargs.update(
        {
            "messages": messages,
            "model": model,
            "max_tokens": max_tokens,
            "temperature": temperature,
            "response_format": response_format,
            "n": n,
        }
    )

    if tools is not None:
        kwargs["tools"] = tools
        if tool_choice not in ["none", "auto", "required"]:
            kwargs["tool_choice"] = {"type": "function", "function": {"name": tool_choice}}
        else:
            kwargs["tool_choice"] = tool_choice

    if seed is not None:
        kwargs["seed"] = seed

    if logprobs is not None:
        kwargs["logprobs"] = logprobs[0]
        if logprobs[0] and logprobs[1] is not None:
            kwargs["top_logprobs"] = logprobs[1]

    start_time = time.time()
    response = client.chat.completions.create(**kwargs)
    end_time = time.time()
    response_duration = end_time - start_time

    response_data: dict[str, Any] = {"choices": []}
    for response_choice in response.choices:
        response_data_curr = {}
        finish_reason = response_choice.finish_reason
        response_data_curr["finish_reason"] = finish_reason

        # We first check for tool calls because even if the finish_reason is stop, the model may have called a tool
        tool_calls = response_choice.message.tool_calls
        if tool_calls:
            tool_names = []
            tool_args_list = []
            for tool_call in tool_calls:
                tool_names.append(tool_call.function.name)
                tool_args_list.append(json.loads(tool_call.function.arguments))
            response_data_curr["message"] = response_choice.message.content
            response_data_curr["tool_names"] = tool_names
            response_data_curr["tool_args_list"] = tool_args_list
        elif finish_reason == "stop" or finish_reason == "length":
            message = response_choice.message.content
            if json_mode:
                with contextlib.suppress(json.JSONDecodeError):
                    message = json.loads(message)
            response_data_curr["message"] = message

        if response_choice.logprobs and response_choice.logprobs.content is not None:
            logprobs_list: list[dict[str, Any] | list[dict[str, Any]]] = []
            for logprob in response_choice.logprobs.content:
                if logprob.top_logprobs:
                    curr_logprob_infos = []
                    for top_logprob in logprob.top_logprobs:
                        curr_logprob_infos.append(
                            {
                                "token": top_logprob.token,
                                "logprob": top_logprob.logprob,
                                "bytes": top_logprob.bytes,
                            }
                        )
                    logprobs_list.append(curr_logprob_infos)
                else:
                    logprobs_list.append(
                        {
                            "token": logprob.token,
                            "logprob": logprob.logprob,
                            "bytes": logprob.bytes,
                        }
                    )

            response_data_curr["logprobs"] = logprobs_list
        response_data["choices"].append(response_data_curr)

    usage = response.usage
    if usage is not None:
        response_data["completion_tokens"] = usage.completion_tokens
        response_data["prompt_tokens"] = usage.prompt_tokens

    if seed is not None and response.system_fingerprint is not None:
        response_data["system_fingerprint"] = response.system_fingerprint

    response_data["response_duration"] = round(response_duration, 4)

    if len(response_data["choices"]) == 1:
        response_data.update(response_data["choices"][0])
        del response_data["choices"]

    return response_data



import contextlib
import json
import re
import time
from typing import Any

from ollama import Client, ResponseError

from not_again_ai.local_llm.ollama.tokens import load_tokenizer, num_tokens_from_messages, num_tokens_in_string


def chat_completion(
    messages: list[dict[str, Any]],
    model: str,
    client: Client,
    max_tokens: int | None = None,
    context_window: int | None = None,
    temperature: float = 0.8,
    json_mode: bool = False,
    seed: int | None = None,
    **kwargs: Any,
) -> dict[str, Any]:
    """Gets a Ollama chat completion response, see https://github.com/ollama/ollama/blob/main/docs/api.md#generate-a-chat-completion
    For a full list of valid parameters: https://github.com/ollama/ollama/blob/main/docs/modelfile.md#valid-parameters-and-values

    Args:
        messages (list[dict[str, Any]]): A list of messages to send to the model.
        model (str): The model to use.
        client (Client): The Ollama client.
        max_tokens (int, optional): The maximum number of tokens to generate. Ollama calls this `num_predict`.
        context_window (int, optional): The number of tokens to consider as context. Ollama calls this `num_ctx`.
        temperature (float, optional): The temperature of the model. Increasing the temperature will make the model answer more creatively.
        json_mode (bool, optional): This will structure the response as a valid JSON object.
            It is important to instruct the model to use JSON in the prompt. Otherwise, the model may generate large amounts whitespace.
        seed (int, optional): The seed to use for the model for reproducible outputs. Defaults to None.

    Returns:
        dict[str, Any]: A dictionary with the following keys
            message (str | dict): The content of the generated assistant message.
                If json_mode is True, this will be a dictionary.
            prompt_tokens (int): The number of tokens in the messages sent to the model.
            completion_tokens (int): The number of tokens used by the model to generate the completion.
            response_duration (float): The time, in seconds, taken to generate the response by using the model.
    """

    options = {
        "num_predict": max_tokens,
        "num_ctx": context_window,
        "temperature": temperature,
    }
    if seed is not None:
        options["seed"] = seed
    options.update(kwargs)

    all_args = {
        "model": model,
        "messages": messages,
        "options": options,
    }
    if json_mode:
        all_args["format"] = "json"

    try:
        start_time = time.time()
        response = client.chat(**all_args)  # type: ignore
        end_time = time.time()
        response_duration = end_time - start_time
    except ResponseError as e:
        # If the error says "model 'model' not found" use regex then raise a more specific error
        expected_pattern = f"model '{model}' not found"
        if re.search(expected_pattern, e.error):
            raise ResponseError(
                f"Model '{model}' not found. Please use not_again_ai.llm.ollama.service.pull() first."
            ) from e
        else:
            raise ResponseError(e.error) from e

    response_data: dict[str, Any] = {}

    # Handle getting the message returned by the model
    message = response["message"].get("content", None)  # type: ignore
    if message and json_mode:
        with contextlib.suppress(json.JSONDecodeError):
            message = json.loads(message)
    if message:
        response_data["message"] = message

    tokenizer = load_tokenizer(model)
    prompt_tokens = num_tokens_from_messages(messages, tokenizer)
    response_data["prompt_tokens"] = prompt_tokens

    # Get the number of tokens generated
    response_data["completion_tokens"] = response.get("eval_count", None)  # type: ignore
    if response_data["completion_tokens"] is None:
        response_data["completion_tokens"] = num_tokens_in_string(str(response_data["message"]), tokenizer)

    # Get the latency of the response
    response_data["response_duration"] = round(response_duration, 4)

    return response_data



from pathlib import Path
from typing import Any

from PIL import Image


def chat_completion_image(
    messages: list[dict[str, str]],
    images: list[Path] | None,
    model_processor: tuple[Any, Any],
    max_tokens: int | None = None,
    temperature: float = 0.7,
) -> dict[str, Any]:
    """A wrapper around ision language model inference for multimodal language models from huggingface.

    Args:
        messages (list[dict[str, str]]): A list of messages to send to the model.
        images (list[Path] | None): A list of image paths to send to the model.
        model_processor (tuple[Any, Any]): A tuple containing the model and processor objects.
        max_tokens (int, optional): The maximum number of tokens to generate. Defaults to None.
        temperature (float, optional): The temperature of the model. Increasing the temperature will make the model answer more creatively. Defaults to 0.7.

    Returns:
        dict[str, Any]: A dictionary with the following keys
            message (str): The content of the generated assistant message.
            completion_tokens (int): The number of tokens used by the model to generate the completion.
    """

    model, processor = model_processor

    prompt = processor.tokenizer.apply_chat_template(messages, tokenize=False, add_generation_prompt=True)

    if images:
        image_objects = [Image.open(image) for image in images]
        inputs = processor(prompt, image_objects, return_tensors="pt").to("cuda:0")
    else:
        inputs = processor(prompt, return_tensors="pt").to("cuda:0")

    generation_args = {
        "max_new_tokens": max_tokens,
        "temperature": temperature,
        "num_beams": 1,
        "do_sample": True,
    }

    generate_ids = model.generate(**inputs, eos_token_id=processor.tokenizer.eos_token_id, **generation_args)

    # Remove input tokens
    generate_ids = generate_ids[:, inputs["input_ids"].shape[1] :]

    # Get the number of generated tokens
    completion_tokens = generate_ids.shape[1]

    response = processor.batch_decode(generate_ids, skip_special_tokens=True, clean_up_tokenization_spaces=False)

    response_data: dict[str, Any] = {}
    response_data["message"] = response[0]
    response_data["completion_tokens"] = completion_tokens
    return response_data



from typing import Any

from transformers import AutoModelForCausalLM, AutoProcessor


def load_model(model_id: str, device_map: str = "cuda", trust_remote_code: bool = True) -> Any:
    """Load a model from Hugging Face."""
    model = AutoModelForCausalLM.from_pretrained(
        model_id,
        device_map=device_map,
        trust_remote_code=trust_remote_code,
        torch_dtype="auto",
    )
    return model


def load_processor(model_id: str, trust_remote_code: bool = True) -> Any:
    """Load a processor from Hugging Face. This is typically used for multimodal language models."""
    processor = AutoProcessor.from_pretrained(
        model_id,
        trust_remote_code=trust_remote_code,
    )
    return processor



import tiktoken
from transformers import AutoTokenizer

from not_again_ai.llm.openai_api.tokens import load_tokenizer as openai_load_tokenizer
from not_again_ai.llm.openai_api.tokens import num_tokens_from_messages as openai_num_tokens_from_messages
from not_again_ai.llm.openai_api.tokens import num_tokens_in_string as openai_num_tokens_in_string
from not_again_ai.llm.openai_api.tokens import truncate_str as openai_truncate_str
from not_again_ai.local_llm.ollama.tokens import load_tokenizer as ollama_load_tokenizer
from not_again_ai.local_llm.ollama.tokens import num_tokens_from_messages as ollama_num_tokens_from_messages
from not_again_ai.local_llm.ollama.tokens import num_tokens_in_string as ollama_num_tokens_in_string
from not_again_ai.local_llm.ollama.tokens import truncate_str as ollama_truncate_str


def load_tokenizer(model: str, provider: str) -> AutoTokenizer | tiktoken.Encoding:
    """Load the tokenizer for the given model and providers

    Args:
        model (str): The name of the language model to load the tokenizer for
        provider (str): Either "openai_api" or "ollama"

    Returns:
        Either a HuggingFace tokenizer or a tiktoken encoding object
    """
    if provider == "openai_api":
        return openai_load_tokenizer(model)
    elif provider == "ollama":
        return ollama_load_tokenizer(model)
    else:
        raise ValueError(f"Unknown tokenizer provider {provider}")


def truncate_str(text: str, max_len: int, tokenizer: AutoTokenizer | tiktoken.Encoding, provider: str) -> str:
    """Truncates a string to a maximum token length.

    Args:
        text: The string to truncate.
        max_len: The maximum number of tokens to keep.
        tokenizer: Either a HuggingFace tokenizer or a tiktoken encoding object
        provider (str): Either "openai_api" or "ollama"

    Returns:
        str: The truncated string.
    """
    if provider == "openai_api":
        return openai_truncate_str(text, max_len, tokenizer)
    elif provider == "ollama":
        return ollama_truncate_str(text, max_len, tokenizer)
    else:
        raise ValueError(f'Unknown tokenizer provider "{provider}"')


def num_tokens_in_string(text: str, tokenizer: AutoTokenizer | tiktoken.Encoding, provider: str) -> int:
    """Return the number of tokens in a string.

    Args:
        text: The string to count the tokens.
        tokenizer: Either a HuggingFace tokenizer or a tiktoken encoding object
        provider (str): Either "openai_api" or "ollama"

    Returns:
        int: The number of tokens in the string.
    """
    if provider == "openai_api":
        return openai_num_tokens_in_string(text, tokenizer)
    elif provider == "ollama":
        return ollama_num_tokens_in_string(text, tokenizer)
    else:
        raise ValueError(f'Unknown tokenizer provider "{provider}"')


def num_tokens_from_messages(
    messages: list[dict[str, str]], tokenizer: AutoTokenizer | tiktoken.Encoding, provider: str
) -> int:
    """Return the number of tokens used by a list of messages.

    Args:
        messages: A list of messages to count the tokens
            should ideally be the result after calling llm.prompts.chat_prompt.
        tokenizer: Either a HuggingFace tokenizer or a tiktoken encoding object
        provider (str): Either "openai_api" or "ollama"

    Returns:
        int: The number of tokens used by the messages.
    """
    if provider == "openai_api":
        return openai_num_tokens_from_messages(messages, tokenizer)
    elif provider == "ollama":
        return ollama_num_tokens_from_messages(messages, tokenizer)
    else:
        raise ValueError(f'Unknown tokenizer provider "{provider}"')



import importlib.util
import os

os.environ["HF_HUB_DISABLE_TELEMETRY"] = "1"
os.environ["TRANSFORMERS_NO_ADVISORY_WARNINGS"] = "1"

if (
    importlib.util.find_spec("liquid") is None
    or importlib.util.find_spec("ollama") is None
    or importlib.util.find_spec("openai") is None
    or importlib.util.find_spec("tiktoken") is None
    or importlib.util.find_spec("transformers") is None
):
    raise ImportError(
        "not_again_ai.local_llm requires the 'llm' and 'local_llm' extra to be installed. "
        "You can install it using 'pip install not_again_ai[llm,local_llm]'."
    )
else:
    import liquid  # noqa: F401
    import ollama  # noqa: F401
    import openai  # noqa: F401
    import tiktoken  # noqa: F401
    import transformers  # noqa: F401
    from transformers.utils import logging

    logging.disable_progress_bar()
    logging.set_verbosity_error()



"""By default use the associated huggingface transformer tokenizer.
If it does not exist in the mapping, default to tiktoken with some buffer (const + percentage)"""

from loguru import logger
import tiktoken
from transformers import AutoTokenizer

from not_again_ai.llm.openai_api.tokens import num_tokens_from_messages as openai_num_tokens_from_messages
from not_again_ai.local_llm.ollama.model_mapping import OLLAMA_MODEL_MAPPING

TIKTOKEN_NUM_TOKENS_BUFFER = 10
TIKTOKEN_PERCENT_TOKENS_BUFFER = 1.1


def load_tokenizer(model: str) -> AutoTokenizer | tiktoken.Encoding:
    """Use the model mapping to load the appropriate tokenizer

    Args:
        model: The name of the language model to load the tokenizer for

    Returns:
        Either a HuggingFace tokenizer or a tiktoken encoding object
    """

    # Loop over the keys in the model mapping checking if the model starts with the key
    for key in OLLAMA_MODEL_MAPPING:
        if model.startswith(key):
            return AutoTokenizer.from_pretrained(OLLAMA_MODEL_MAPPING[key], use_fast=True)

    # If the model does not start with any key in the model mapping, default to tiktoken
    logger.warning(
        f'Model "{model}" not found in OLLAMA_MODEL_MAPPING. Using tiktoken - token counts will have an added buffer of \
{TIKTOKEN_PERCENT_TOKENS_BUFFER * 100}% plus {TIKTOKEN_NUM_TOKENS_BUFFER} tokens.'
    )
    tokenizer = tiktoken.get_encoding("o200k_base")
    return tokenizer


def truncate_str(text: str, max_len: int, tokenizer: AutoTokenizer | tiktoken.Encoding) -> str:
    """Truncates a string to a maximum token length.

    Args:
        text: The string to truncate.
        max_len: The maximum number of tokens to keep.
        tokenizer: Either a HuggingFace tokenizer or a tiktoken encoding object

    Returns:
        str: The truncated string.
    """
    if isinstance(tokenizer, tiktoken.Encoding):
        tokens = tokenizer.encode(text)
        if len(tokens) > max_len:
            tokens = tokens[:max_len]
            truncated_text = tokenizer.decode(tokens)
            return truncated_text
    else:
        tokens = tokenizer(text, return_tensors=None)["input_ids"]
        if len(tokens) > max_len:
            tokens = tokens[:max_len]
            truncated_text = tokenizer.decode(tokens)
            return truncated_text

    return text


def num_tokens_in_string(text: str, tokenizer: AutoTokenizer | tiktoken.Encoding) -> int:
    """Return the number of tokens in a string.

    Args:
        text: The string to count the tokens.
        tokenizer: Either a HuggingFace tokenizer or a tiktoken encoding object

    Returns:
        int: The number of tokens in the string.
    """
    if isinstance(tokenizer, tiktoken.Encoding):
        num_tokens = (len(tokenizer.encode(text)) * TIKTOKEN_PERCENT_TOKENS_BUFFER) + TIKTOKEN_NUM_TOKENS_BUFFER
        return int(num_tokens)
    else:
        tokens = tokenizer(text, return_tensors=None)["input_ids"]
        return len(tokens)


def num_tokens_from_messages(messages: list[dict[str, str]], tokenizer: AutoTokenizer | tiktoken.Encoding) -> int:
    """Return the number of tokens used by a list of messages.
    For models with HuggingFace tokenizers, uses

    Args:
        messages: A list of messages to count the tokens
            should ideally be the result after calling llm.prompts.chat_prompt.
        tokenizer: Either a HuggingFace tokenizer or a tiktoken encoding object

    Returns:
        int: The number of tokens used by the messages.
    """
    if isinstance(tokenizer, tiktoken.Encoding):
        num_tokens = (
            openai_num_tokens_from_messages(messages, tokenizer=tokenizer, model="gpt-4o")
            * TIKTOKEN_PERCENT_TOKENS_BUFFER
        ) + TIKTOKEN_NUM_TOKENS_BUFFER
        return int(num_tokens)
    else:
        tokens = tokenizer.apply_chat_template(messages, add_generation_prompt=True, return_tensors=None)
        return len(tokens)



"""Hardcoded mapping from ollama model names to their associated HuggingFace tokenizer.

Given the way that Ollama models are tagged, we can against the first part of the model name,
i.e. all phi3 models will start with "phi3".
"""

OLLAMA_MODEL_MAPPING = {
    "phi3": "microsoft/Phi-3-mini-4k-instruct",
    "llama3:": "nvidia/Llama3-ChatQA-1.5-8B",  # Using this version to get around needed to accept an agreement to get access to the tokenizer
    "gemma": "google/gemma-1.1-7b-it",  # Requires HF_TOKEN set and accepting the agreement on the HF model page
    "qwen2": "Qwen/Qwen2-7B-Instruct",
    "granite-code": "ibm-granite/granite-34b-code-instruct",
    "llama3-gradient": "nvidia/Llama3-ChatQA-1.5-8B",
    "command-r": "CohereForAI/c4ai-command-r-v01",
    "deepseek-coder-v2": "deepseek-ai/DeepSeek-Coder-V2-Lite-Base",
}



import os

from ollama import Client


def ollama_client(host: str | None = None, timeout: float | None = None) -> Client:
    """Create an Ollama client instance based on the specified host or will read from the OLLAMA_HOST environment variable.

    Args:
        host (str, optional): The host URL of the Ollama server.
        timeout (float, optional): The timeout for requests

    Returns:
        Client: An instance of the Ollama client.

    Examples:
        >>> client = client(host="http://localhost:11434")
    """
    if host is None:
        host = os.getenv("OLLAMA_HOST")
        if host is None:
            raise ValueError("Host must be provided or OLLAMA_HOST environment variable must be set.")

    return Client(host=host, timeout=timeout)



from copy import deepcopy

from liquid import Template


def chat_prompt(messages_unformatted: list[dict[str, str]], variables: dict[str, str]) -> list[dict[str, str]]:
    """Formats a list of messages for chat completion models using Liquid templating.

    Args:
        messages_unformatted: A list of dictionaries where each dictionary
            represents a message. Each message must have 'role' and 'content'
            keys with string values, where content is a Liquid template.
        variables: A dictionary where each key-value pair represents a variable
            name and its value for template rendering.

    Returns:
        A list of dictionaries with the same structure as `messages_unformatted`,
        but with the 'content' of each message with the provided `variables`.

    Examples:
        >>> messages = [
        ...     {"role": "system", "content": "You are a helpful assistant."},
        ...     {"role": "user", "content": "Help me {{task}}"}
        ... ]
        >>> vars = {"task": "write Python code for the fibonnaci sequence"}
        >>> chat_prompt(messages, vars)
        [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "Help me write Python code for the fibonnaci sequence"}
        ]
    """

    messages_formatted = deepcopy(messages_unformatted)
    for message in messages_formatted:
        liquid_template = Template(message["content"])
        message["content"] = liquid_template.render(**variables)

    return messages_formatted
